﻿Imports System.IO
Imports System.Text.RegularExpressions
Imports Newtonsoft.Json

Public Class Form1

    ' Class to store address entries
    Public Class AddressEntry
        Public Property Name As String
        Public Property Address As String
        Public Property Phone As String
        Public Property Category As String
        Public Property DateAdded As DateTime

        Public Overrides Function ToString() As String
            Return $"{Name}, {Address}, {Phone}"
        End Function

        ' Constructor
        Public Sub New(name As String, address As String, phone As String, category As String)
            Me.Name = name
            Me.Address = address
            Me.Phone = phone
            Me.Category = category
            Me.DateAdded = DateTime.Now
        End Sub
    End Class

    ' List to hold address entries
    Private addressList As New List(Of AddressEntry)
    Private Const FilePath As String = "addresses.json"
    Private caseSensitive As Boolean = False
    Private matchWholeWords As Boolean = False
    Private searchHistory As New List(Of String)()
    Private isEdit As Boolean = False
    Private synonymDictionary As New Dictionary(Of String, List(Of String))()

    ' Create a dictionary to hold synonyms
    ''Private synonymDictionary As New Dictionary(Of String, List(Of String)) From {
    ''{"quick", New List(Of String) From {"fast", "swift", "speedy"}},
    ''{"jumps", New List(Of String) From {"leaps", "bounds"}},
    ''{"lazy", New List(Of String) From {"idle", "sluggish"}}
    ''}

    ' Load addresses from file on startup
    Private Sub AddressManager_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        LoadAddresses()
        cmbCategoryFilter.Items.AddRange(New String() {"Family", "Friends", "Work", "Others"})
        LoadSynonymsFromFile("synonyms.txt")
        ''lstSuggestions.Visible = False ' Ensure it starts hidden
    End Sub

    ' Optional: hide suggestions when clicking elsewhere
    Private Sub AddressManager_Click(sender As Object, e As EventArgs) Handles MyBase.Click
        ''lstSuggestions.Visible = False
    End Sub

    ' Event handler for Add button
    Private Sub btnAdd_Click(sender As Object, e As EventArgs) Handles btnAdd.Click
        If ValidateFields() Then
            Dim newEntry As New AddressEntry(txtName.Text, txtAddress.Text, txtPhone.Text, cmbCategoryFilter.SelectedItem.ToString)
            addressList.Add(newEntry)
            isEdit = False
            UpdateAddressListBox()
            ClearFields()
            MessageBox.Show("Entry Added Successfully!")
        End If
    End Sub

    Private Sub ChangeControls(isEnabled As Boolean)
        btnEdit.Enabled = isEnabled
        btnDelete.Enabled = isEnabled
        btnExport.Enabled = isEnabled
        btnImport.Enabled = isEnabled
        btnSortByName.Enabled = isEnabled
        btnSortByPhone.Enabled = isEnabled
        btnFilterByDate.Enabled = isEnabled
        btnSearch.Enabled = isEnabled
        btnClearSearch.Enabled = isEnabled
        btnClearSearchHistory.Enabled = isEnabled
        btnSettings.Enabled = isEnabled
        dtpStartDate.Enabled = isEnabled
        dtpEndDate.Enabled = isEnabled
        txtSearch.Enabled = isEnabled
    End Sub

    ' Event handler for Edit button
    Private Sub btnEdit_Click(sender As Object, e As EventArgs) Handles btnEdit.Click
        If lstAddresses.SelectedIndex >= 0 Then
            ChangeControls(False)
            isEdit = True
            Dim selectedIndex = lstAddresses.SelectedIndex
            Dim selectedEntry = addressList(selectedIndex)

            txtName.Text = selectedEntry.Name
            txtAddress.Text = selectedEntry.Address
            txtPhone.Text = selectedEntry.Phone
            cmbCategoryFilter.SelectedItem = selectedEntry.Category ' Set the selected category here

            ' Remove the entry temporarily while editing
            addressList.RemoveAt(selectedIndex)
            UpdateAddressListBox()
        Else
            MessageBox.Show("Please select an entry to edit.")
        End If
    End Sub

    Private Sub btnDelete_Click(sender As Object, e As EventArgs) Handles btnDelete.Click
        If lstAddresses.SelectedIndex >= 0 Then
            Dim result = MessageBox.Show("Are you sure you want to delete this entry?", "Confirm Delete", MessageBoxButtons.YesNo)
            If result = DialogResult.Yes Then
                addressList.RemoveAt(lstAddresses.SelectedIndex)
                UpdateAddressListBox()
                MessageBox.Show("Entry Deleted Successfully!")
            End If
        Else
            MessageBox.Show("Please select an entry to delete.")
        End If
    End Sub

    ' Event handler for Save File on Form Closing
    Private Sub AddressManager_FormClosing(sender As Object, e As FormClosingEventArgs) Handles Me.FormClosing
        If isEdit = True Then
            Dim newEntry As New AddressEntry(txtName.Text, txtAddress.Text, txtPhone.Text, cmbCategoryFilter.SelectedItem.ToString)
            addressList.Add(newEntry)
        End If
        SaveAddresses()
    End Sub

    ' Function to validate fields
    Private Function ValidateFields() As Boolean
        If String.IsNullOrWhiteSpace(txtName.Text) Or String.IsNullOrWhiteSpace(txtAddress.Text) Or String.IsNullOrWhiteSpace(txtPhone.Text) Or cmbCategoryFilter.SelectedItem Is Nothing Then
            MessageBox.Show("Please fill in all fields.")
            Return False
        End If
        Return True
    End Function

    ' Function to update ListBox with current addresses
    Private Sub UpdateAddressListBox(Optional ByVal filteredEntries As List(Of AddressEntry) = Nothing)
        lstAddresses.Items.Clear()
        Dim entriesToDisplay As List(Of AddressEntry)

        If filteredEntries IsNot Nothing Then
            entriesToDisplay = filteredEntries
        Else
            entriesToDisplay = addressList
        End If

        For Each entry In entriesToDisplay
            Dim searchTerm As String = txtSearch.Text.Trim() ' Get the current search term

            If searchTerm <> "" Then
                Dim highlightedText As String = HighlightMatches(entry, searchTerm)
                lstAddresses.Items.Add(highlightedText)
            Else
                lstAddresses.Items.Add(entry)
            End If
        Next

    End Sub

    ' Function to clear input fields
    Private Sub ClearFields()
        txtName.Clear()
        txtAddress.Clear()
        txtPhone.Clear()
        cmbCategoryFilter.SelectedIndex = -1
    End Sub

    ' Function to load addresses from file
    Private Sub LoadAddresses()
        If File.Exists(FilePath) Then
            Dim jsonData = File.ReadAllText(FilePath)
            addressList = JsonConvert.DeserializeObject(Of List(Of AddressEntry))(jsonData)
            UpdateAddressListBox()
        End If
    End Sub

    ' Function to save addresses to file
    Private Sub SaveAddresses()
        Dim jsonData = JsonConvert.SerializeObject(addressList, Formatting.Indented)
        File.WriteAllText(FilePath, jsonData)
    End Sub

    Private Sub btnSearch_Click(sender As Object, e As EventArgs) Handles btnSearch.Click
        Dim searchTerm As String = txtSearch.Text.Trim()

        If Not String.IsNullOrWhiteSpace(searchTerm) Then
            ' Add to search history if not already present
            If Not searchHistory.Contains(searchTerm) Then
                searchHistory.Add(searchTerm)
                UpdateSearchHistoryDropdown()
            End If

            ' Get search terms including synonyms
            Dim searchTerms As List(Of String) = GetSearchTermsIncludingSynonyms(searchTerm)
            Dim searchEntries As New List(Of AddressEntry)()

            ' Loop through each address entry in the address list
            For Each entry As AddressEntry In addressList
                Dim matches As Boolean = False

                ' Prepare the entry properties for comparison
                Dim entryName As String = If(caseSensitive, entry.Name, entry.Name.ToLower())
                Dim entryAddress As String = If(caseSensitive, entry.Address, entry.Address.ToLower())
                Dim entryPhone As String = If(caseSensitive, entry.Phone, entry.Phone.ToLower())

                ' Check for each term in the search terms
                For Each term As String In searchTerms
                    Dim termToMatch As String = If(caseSensitive, term, term.ToLower())

                    ' Check for fuzzy match using Levenshtein distance
                    If Levenshtein(entryName, termToMatch) <= 1 OrElse
                   Levenshtein(entryAddress, termToMatch) <= 1 OrElse
                   Levenshtein(entryPhone, termToMatch) <= 1 Then
                        matches = True
                        Exit For
                    End If

                    ' Use regex for whole word matching if required
                    If matchWholeWords Then
                        Dim pattern As String = $"\b{Regex.Escape(termToMatch)}\b"
                        If Regex.IsMatch(entryName, pattern) OrElse
                       Regex.IsMatch(entryAddress, pattern) OrElse
                       Regex.IsMatch(entryPhone, pattern) Then
                            matches = True
                            Exit For
                        End If
                    Else
                        ' Regular sub-string matching (with case sensitivity check)
                        If entryName.IndexOf(termToMatch) >= 0 OrElse
                       entryAddress.IndexOf(termToMatch) >= 0 OrElse
                       entryPhone.IndexOf(termToMatch) >= 0 Then
                            matches = True
                            Exit For
                        End If
                    End If
                Next

                ' If any of the search terms matched, add to the result list
                If matches Then
                    searchEntries.Add(entry)
                End If
            Next

            ' Sort the matching entries by name
            searchEntries.Sort(Function(a, b) String.Compare(a.Name, b.Name))

            ' Display the results or show a message if no matches found
            If searchEntries.Count > 0 Then
                UpdateAddressListBox(searchEntries)
            Else
                MessageBox.Show("No matching entries found.", "Search Result", MessageBoxButtons.OK, MessageBoxIcon.Information)
                UpdateAddressListBox(addressList)
            End If
        Else
            UpdateAddressListBox(addressList) ' Show full list if search term is empty
        End If
    End Sub

    ' Event handler for Category filter change
    Private Sub cmbCategoryFilter_SelectedIndexChanged(sender As Object, e As EventArgs) Handles cmbCategoryFilter.SelectedIndexChanged
        If cmbCategoryFilter.SelectedItem IsNot Nothing Then
            Dim selectedCategory = cmbCategoryFilter.SelectedItem.ToString()
            Dim filteredEntries = addressList.Where(Function(entry) entry.Category = selectedCategory).ToList()
            UpdateAddressListBox(filteredEntries)
        Else
            UpdateAddressListBox()
        End If
    End Sub

    Private Sub btnSortByName_Click(sender As Object, e As EventArgs) Handles btnSortByName.Click
        addressList = addressList.OrderBy(Function(entry) entry.Name).ToList()
        UpdateAddressListBox()
    End Sub

    Private Sub btnSortByPhone_Click(sender As Object, e As EventArgs) Handles btnSortByPhone.Click
        addressList = addressList.OrderBy(Function(entry) entry.Phone).ToList()
        UpdateAddressListBox()
    End Sub

    Private Sub btnExport_Click(sender As Object, e As EventArgs) Handles btnExport.Click
        Dim saveFileDialog As New SaveFileDialog With {
            .Filter = "CSV Files (*.csv)|*.csv"
        }
        If saveFileDialog.ShowDialog() = DialogResult.OK Then
            Using writer As New StreamWriter(saveFileDialog.FileName)
                For Each entry In addressList
                    writer.WriteLine($"{entry.Name},{entry.Address},{entry.Phone},{entry.Category}")
                Next
            End Using
        End If
    End Sub

    Private Sub btnImport_Click(sender As Object, e As EventArgs) Handles btnImport.Click
        Dim openFileDialog As New OpenFileDialog With {
            .Filter = "CSV Files (*.csv)|*.csv"
        }
        If openFileDialog.ShowDialog() = DialogResult.OK Then
            Dim lines = File.ReadAllLines(openFileDialog.FileName)
            For Each line In lines
                Dim parts = line.Split(","c)
                If parts.Length = 4 Then
                    addressList.Add(New AddressEntry(parts(0), parts(1), parts(2), parts(3)))
                End If
            Next
            UpdateAddressListBox()
        End If
    End Sub

    ' Clear search functionality
    Private Sub btnClearSearch_Click(sender As Object, e As EventArgs) Handles btnClearSearch.Click
        txtSearch.Text = String.Empty
        UpdateAddressListBox(addressList.ToList())
    End Sub

    ' Highlight matches function (basic implementation)
    Private Function HighlightMatches(entry As AddressEntry, searchTerm As String) As String
        ' Basic example: highlighting matches by wrapping them in asterisks
        Dim name As String = entry.Name.Replace(searchTerm, $"*{searchTerm}*")
        Dim address As String = entry.Address.Replace(searchTerm, $"*{searchTerm}*")
        Dim phone As String = entry.Phone.Replace(searchTerm, $"*{searchTerm}*")
        Return $"{name}, {address}, {phone}"
    End Function

    ' Add this method to open settings
    Private Sub btnSettings_Click(sender As Object, e As EventArgs) Handles btnSettings.Click
        Dim settingsForm As New SettingsForm
        settingsForm.IsCaseSensitive = caseSensitive
        settingsForm.MatchWholeWords = matchWholeWords

        If settingsForm.ShowDialog() = DialogResult.OK Then
            caseSensitive = settingsForm.IsCaseSensitive
            matchWholeWords = settingsForm.MatchWholeWords
        End If
    End Sub

    Private Sub UpdateSearchHistoryDropdown()
        cmbSearchHistory.Items.Clear() ' Clear the existing items

        For Each term In searchHistory
            cmbSearchHistory.Items.Add(term)
        Next
    End Sub

    Private Sub cmbSearchHistory_SelectedIndexChanged(sender As Object, e As EventArgs) Handles cmbSearchHistory.SelectedIndexChanged
        If cmbSearchHistory.SelectedItem IsNot Nothing Then
            txtSearch.Text = cmbSearchHistory.SelectedItem.ToString()
        End If
    End Sub

    Private Sub btnClearSearchHistory_Click(sender As Object, e As EventArgs) Handles btnClearSearchHistory.Click
        searchHistory.Clear()
        cmbSearchHistory.Items.Clear() ' Clear the dropdown as well
        cmbSearchHistory.Text = ""
    End Sub

    ' Get relevant search terms using synonyms
    Private Function GetSearchTermsIncludingSynonyms(searchTerm As String) As List(Of String)
        Dim searchTerms As New List(Of String) From {searchTerm}
        Dim lowerTerm As String = searchTerm.ToLower()

        If synonymDictionary.ContainsKey(lowerTerm) Then
            searchTerms.AddRange(synonymDictionary(lowerTerm))
        End If

        Return searchTerms
    End Function

    ' Function to calculate Levenshtein Distance
    Private Function Levenshtein(s1 As String, s2 As String) As Integer
        Dim d(s1.Length, s2.Length) As Integer

        For i As Integer = 0 To s1.Length
            d(i, 0) = i
        Next
        For j As Integer = 0 To s2.Length
            d(0, j) = j
        Next

        For i As Integer = 1 To s1.Length
            For j As Integer = 1 To s2.Length
                Dim cost As Integer = If(s1(i - 1) = s2(j - 1), 0, 1)
                d(i, j) = Math.Min(Math.Min(d(i - 1, j) + 1, d(i, j - 1) + 1), d(i - 1, j - 1) + cost)
            Next
        Next

        Return d(s1.Length, s2.Length)
    End Function

    Private Sub LoadSynonymsFromFile(filePath As String)
        If Not File.Exists(Application.StartupPath + "\" + filePath) Then
            MessageBox.Show("Synonyms file not found.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
            Return
        End If

        ' Read the file line by line
        Dim lines() As String = File.ReadAllLines(filePath)

        For Each line As String In lines
            ' Split the line by comma
            Dim parts As String() = line.Split(","c)

            ' Check cases depending on the length of parts
            If parts.Length = 0 Then
                ' No content in line
                Continue For
            ElseIf parts.Length = 1 Then
                ' Only one part (no synonyms provided), we could log this or handle it accordingly
                Console.WriteLine($"Warning: No synonyms provided for '{parts(0).Trim()}'. Skipping line.")
                Continue For
            Else
                ' More than one part, we can take the first as the word and the rest as synonyms
                Dim word As String = parts(0).Trim().ToLower()
                Dim synonyms As New List(Of String)()

                ' Loop through the remaining parts to add synonyms
                For i As Integer = 1 To parts.Length - 1
                    Dim synonym As String = parts(i).Trim()
                    If Not String.IsNullOrWhiteSpace(synonym) Then
                        synonyms.Add(synonym) ' Add only if it's not empty
                    End If
                Next

                ' Add to the dictionary
                If Not synonymDictionary.ContainsKey(word) Then
                    synonymDictionary(word) = New List(Of String)()
                End If
                synonymDictionary(word).AddRange(synonyms) ' Add all synonyms to the dictionary
            End If
        Next
    End Sub

    ' Add this in your existing Form1 class
    Private Sub txtSearch_TextChanged(sender As Object, e As EventArgs) Handles txtSearch.TextChanged
        Dim searchTerm As String = txtSearch.Text.Trim()

        If Not String.IsNullOrWhiteSpace(searchTerm) Then
            ' Get relevant suggestions
            Dim suggestions = GetSuggestions(searchTerm)
            UpdateSuggestionsList(suggestions)
        Else
            ' Clear suggestions if the search term is empty
            lstSuggestions.Items.Clear()
            ''lstSuggestions.Visible = False
        End If
    End Sub

    ' Function to get search suggestions based on input
    Private Function GetSuggestions(searchTerm As String) As List(Of String)
        Dim suggestions As New List(Of String)

        For Each entry In addressList
            If entry.Name.ToLower().Contains(searchTerm.ToLower()) Then
                suggestions.Add(entry.Name)
            End If
            If entry.Address.ToLower().Contains(searchTerm.ToLower()) Then
                suggestions.Add(entry.Address)
            End If
            If entry.Phone.ToLower().Contains(searchTerm.ToLower()) Then
                suggestions.Add(entry.Phone)
            End If
        Next

        Return suggestions
    End Function

    ' Function to update suggestions in the ListBox
    Private Sub UpdateSuggestionsList(suggestions As List(Of String))
        lstSuggestions.Items.Clear()

        For Each suggestion In suggestions
            lstSuggestions.Items.Add(suggestion)
        Next

        ''lstSuggestions.Visible = suggestions.Count > 0 ' Show the list if there are suggestions
    End Sub

    ' Double-click event to fill the textbox with the selected suggestion
    Private Sub lstSuggestions_DoubleClick(sender As Object, e As EventArgs) Handles lstSuggestions.DoubleClick
        If lstSuggestions.SelectedIndex <> -1 Then
            txtSearch.Text = lstSuggestions.SelectedItem.ToString()
            ''lstSuggestions.Visible = False ' Hide the suggestions list after selection
        End If
    End Sub

    Private Sub btnFilterByDate_Click(sender As Object, e As EventArgs) Handles btnFilterByDate.Click
        Dim startDate As DateTime = dtpStartDate.Value
        Dim endDate As DateTime = dtpEndDate.Value

        Dim filteredEntries = addressList.Where(Function(entry) entry.DateAdded >= startDate AndAlso entry.DateAdded <= endDate).ToList()
        UpdateAddressListBox(filteredEntries)
    End Sub

End Class
